@@ -18,7 +18,7 @@
  * Rights Reserved.
  *
  * Contributor(s):
- *   Jan Varga (varga@utcru.sk)
+ *   Jan Varga (varga@nixcorp.com)
  *   Hkan Waara (hwaara@chello.se)
  *   Neil Rashbrook (neil@parkwaycc.co.uk)
  *   Seth Spitzer <sspitzer@netscape.com>
@@ -60,6 +60,7 @@ var gRightMouseButtonDown = false;
 // This is used to make sure that the row with the currentIndex has the selection
 // after a Delete or Move of a message that has a row index less than currentIndex.
 var gThreadPaneCurrentSelectedIndex = -1;
+var gLoadStartFolder = true;
 
 // Global var to keep track of if the 'Delete Message' or 'Move To' thread pane
 // context menu item was triggered.  This helps prevent the tree view from
@@ -149,13 +150,8 @@ var folderListener = {
                if (gRerootOnFolderLoad)
                  RerootFolder(uri, msgFolder, gCurrentLoadingFolderViewType, gCurrentLoadingFolderViewFlags, gCurrentLoadingFolderSortType, gCurrentLoadingFolderSortOrder);
 
-               var db = msgFolder.getMsgDatabase(msgWindow);
-               if (db) 
-                 db.resetHdrCacheSize(100);
-                 
-               if (gDBView) {
+               if (gDBView) 
                  gDBView.suppressCommandUpdating = false;
-               }
 
                gIsEditableMsgFolder = IsSpecialFolder(msgFolder, MSG_FOLDER_FLAG_DRAFTS);
 
@@ -254,6 +250,9 @@ var folderListener = {
            {
              Search("");
            }
+           else {
+             ViewChangeByValue(pref.getIntPref("mailnews.view.last"));
+           }
          }
        } 
        else if (eventType == "ImapHdrDownloaded") {
@@ -624,7 +623,6 @@ function OnLoadMessenger()
   verifyAccounts(null);
     
   HideAccountCentral();
-  loadStartPage();
   InitMsgWindow();
   messenger.SetWindow(window, msgWindow);
 
@@ -640,18 +638,14 @@ function OnLoadMessenger()
   // argument[0] --> folder uri
   // argument[1] --> optional message key
   // argument[2] --> optional email address; //will come from aim; needs to show msgs from buddy's email address  
-  if ("arguments" in window && window.arguments[0])
+  if ("arguments" in window)
   {
-    gStartFolderUri = window.arguments[0];
-    gStartMsgKey = window.arguments[1];
-    gSearchEmailAddress = window.arguments[2];
-
-  }
-  else
-  {
-    gStartFolderUri = null;
-    gStartMsgKey = nsMsgKey_None;
-    gSearchEmailAddress = null;
+    gStartFolderUri = (window.arguments.length > 0) ? window.arguments[0]
+                                                    : null;
+    gStartMsgKey = (window.arguments.length > 1) ? window.arguments[1]
+                                                 : nsMsgKey_None;
+    gSearchEmailAddress = (window.arguments.length > 2) ? window.arguments[2]
+                                                        : null;
   }
 
   setTimeout("loadStartFolder(gStartFolderUri);", 0);
@@ -671,6 +665,29 @@ function OnLoadMessenger()
     messengerBox.setAttribute("width","500px");
   }
 
+  if (CheckForUnsentMessages())
+  {
+    var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
+    if (!ioService.offline) 
+    {
+      InitPrompts();
+      InitServices();
+
+      if (gPromptService) 
+      {
+        var buttonPressed = gPromptService.confirmEx(window, 
+                            gOfflinePromptsBundle.getString('sendMessagesOfflineWindowTitle'), 
+                            gOfflinePromptsBundle.getString('sendMessagesLabel'),
+                            gPromptService.BUTTON_TITLE_IS_STRING * (gPromptService.BUTTON_POS_0 + 
+                              gPromptService.BUTTON_POS_1),
+                            gOfflinePromptsBundle.getString('sendMessagesSendButtonLabel'),
+                            gOfflinePromptsBundle.getString('sendMessagesNoSendButtonLabel'),
+                            null, null, {value:0});
+        if (buttonPressed == 0) 
+          SendUnsentMessages();
+      }
+    }
+  }
 
   //Set focus to the Thread Pane the first time the window is opened.
   SetFocusThreadPane();
@@ -775,16 +792,17 @@ function loadStartFolder(initialUri)
             }
         }
 
-        var startFolder = startFolderResource.QueryInterface(Components.interfaces.nsIFolder);
+        var startFolder = startFolderResource.QueryInterface(Components.interfaces.nsIMsgFolder);
         SelectFolder(startFolder.URI);
-                
+
         // only do this on startup, when we pass in null
-        if (!initialUri && isLoginAtStartUpEnabled)
+        if (!initialUri && isLoginAtStartUpEnabled && gLoadStartFolder)
         {
             // Perform biff on the server to check for new mail, except for imap
             if (defaultServer.type != "imap")
-              defaultServer.PerformBiff();         
-        } 
+              defaultServer.PerformBiff(msgWindow);        
+        }
+
 
         // because the "open" state persists, we'll call
         // PerformExpand() for all servers that are open at startup.
@@ -919,10 +937,37 @@ function UpgradeThreadPaneUI()
 
   try {
     threadPaneUIVersion = pref.getIntPref("mailnews.ui.threadpane.version");
-    if (threadPaneUIVersion == 1) {
-      labelCol = document.getElementById("labelCol");
-      labelCol.setAttribute("hidden", "true");
-      pref.setIntPref("mailnews.ui.threadpane.version", 2);
+    if (threadPaneUIVersion < 4) {
+      var threadTree = document.getElementById("threadTree");
+      var junkCol = document.getElementById("junkStatusCol");
+      var beforeCol;
+
+      if (threadPaneUIVersion < 3) {
+        var subjectCol = document.getElementById("subjectCol");
+      
+        beforeCol = subjectCol.boxObject.nextSibling.boxObject.nextSibling;
+        if (beforeCol)
+          threadTree._reorderColumn(junkCol, beforeCol, true);
+        else // subjectCol was the last column, put it after
+          threadTree._reorderColumn(junkCol, subjectCol, false);
+
+        if (threadPaneUIVersion == 1) {
+          labelCol = document.getElementById("labelCol");
+          labelCol.setAttribute("hidden", "true");
+        }
+      }
+
+      var senderCol = document.getElementById("senderCol");
+      var recipientCol = document.getElementById("recipientCol");
+    
+      beforeCol = junkCol.boxObject.nextSibling.boxObject.nextSibling;
+      if (beforeCol)
+        threadTree._reorderColumn(recipientCol, beforeCol, true);
+      else // junkCol was the last column, put it after
+        threadTree._reorderColumn(recipientCol, junkCol, false);
+      threadTree._reorderColumn(senderCol, recipientCol, true);
+
+      pref.setIntPref("mailnews.ui.threadpane.version", 4);
     }
 	}
   catch (ex) {
@@ -1352,15 +1397,17 @@ function SetNextMessageAfterDelete()
 {
   var treeSelection = GetThreadTree().treeBoxObject.selection;
 
-  // Only set gThreadPaneDeleteOrMoveOccurred to true if the message was
-  // truly moved to the trash or deleted, as opposed to an IMAP delete
-  // (where it is only "marked as deleted".  This will prevent bug 142065.
-  //
-  // If it's an IMAP delete, then just set gNextMessageViewIndexAfterDelete
-  // to treeSelection.currentIndex (where the outline is at) because nothing
-  // was moved or deleted from the folder.
-  if(gDBView.removeRowOnMoveOrDelete)
+  if (treeSelection.isSelected(treeSelection.currentIndex))
+    gNextMessageViewIndexAfterDelete = gDBView.msgToSelectAfterDelete;
+  else if(gDBView.removeRowOnMoveOrDelete)
   {
+    // Only set gThreadPaneDeleteOrMoveOccurred to true if the message was
+    // truly moved to the trash or deleted, as opposed to an IMAP delete
+    // (where it is only "marked as deleted".  This will prevent bug 142065.
+    //
+    // If it's an IMAP delete, then just set gNextMessageViewIndexAfterDelete
+    // to treeSelection.currentIndex (where the outline is at) because nothing
+    // was moved or deleted from the folder.
     gThreadPaneDeleteOrMoveOccurred = true;
     gNextMessageViewIndexAfterDelete = treeSelection.currentIndex - NumberOfSelectedMessagesAboveCurrentIndex(treeSelection.currentIndex);
   }
@@ -1368,42 +1415,28 @@ function SetNextMessageAfterDelete()
     gNextMessageViewIndexAfterDelete = treeSelection.currentIndex;
 }
 
-function EnsureAllAncestorsAreExpanded(tree, resource)
+function EnsureFolderIndex(builder, msgFolder)
 {
-    // get the parent of the desired folder, and then try to get
-    // the index of the parent in the tree
-    var folder = resource.QueryInterface(Components.interfaces.nsIFolder);
-    
-    // if this is a server, there are no ancestors, so stop.
-    var msgFolder = folder.QueryInterface(Components.interfaces.nsIMsgFolder);
-    if (msgFolder.isServer)
-      return;
-
-    var parentFolderResource = RDF.GetResource(folder.parent.URI);
-    var folderIndex = GetFolderIndex(tree, parentFolderResource);
-
-    if (folderIndex == -1) {
-      // if we couldn't find the parent, recurse
-      EnsureAllAncestorsAreExpanded(tree, parentFolderResource);
-      // ok, now we should be able to find the parent
-      folderIndex = GetFolderIndex(tree, parentFolderResource);
-    }
-
-    // if the parent isn't open, open it
-    if (!(tree.treeBoxObject.view.isContainerOpen(folderIndex)))
-      tree.treeBoxObject.view.toggleOpenState(folderIndex);
+  // try to get the index of the folder in the tree
+  var index = builder.getIndexOfResource(msgFolder);
+  if (index == -1) {
+    // if we couldn't find the folder, open the parent
+    builder.toggleOpenState(EnsureFolderIndex(builder, msgFolder.parent));
+    index = builder.getIndexOfResource(msgFolder);
+  }
+  return index;
 }
 
 function SelectFolder(folderUri)
 {
     var folderTree = GetFolderTree();
     var folderResource = RDF.GetResource(folderUri);
+    var msgFolder = folderResource.QueryInterface(Components.interfaces.nsIMsgFolder);
 
     // before we can select a folder, we need to make sure it is "visible"
     // in the tree.  to do that, we need to ensure that all its
     // ancestors are expanded
-    EnsureAllAncestorsAreExpanded(folderTree, folderResource);
-    var folderIndex = GetFolderIndex(folderTree, folderResource);
+    var folderIndex = EnsureFolderIndex(folderTree.builderView, msgFolder);
     ChangeSelection(folderTree, folderIndex);
 }
 
